//////////////////////////////////////////////////////////////////////
// Tom and Jerry simple failure tests for formats
// Gavin Brown, Michael Kerber, Miles Reid, 2005-10
//
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
//		Test Tom and Jerry matrices
//////////////////////////////////////////////////////////////////////

// First, a map R -> R (= poly ring) that sets all vars to 0 except R.i.
function set_to_zero_except(R,i)
    return hom< R -> R | [ R | 0 : j in [1..i-1] ] cat [ R.i ] cat
        [ R | 0 : j in [i+1..Rank(R)] ] >;
end function;

// ... and a map R -> R (= poly ring) that sets all vars to 0 except R.i -> 1.
function set_to_zero_except_i_to_one(R,i)
    return hom< R -> R | [ R | 0 : j in [1..i-1] ] cat [ 1 ] cat
        [ R | 0 : j in [i+1..Rank(R)] ] >;
end function;

// Two cheap tests on the appearance of variables in the eqns/matrix
// and a third similar but more fiddly.

// 1:
// Y has a bad cone point if a variable doesn't appear in the eqns at all.
function does_each_variable_appear(IY)
        return [ &or[ Degree(f,RR.i) gt 0 : f in Basis(IY) ] :
			i in [1..Rank(RR)] ]
			where RR := Generic(IY);
end function;

// 2:
// Y has a non-terminal point if a variable doesn't appear as a pure power
// somewhere in the matrix since that point has embedding dimension 6.
function point_of_embdim6(M)
        return [ &or[ set_to_zero_except(RR,i)(m) ne 0 : m in Eltseq(M) ] :
			i in [1..Rank(RR)] ]
			where RR := Parent(M[1,1]);
end function;

/*
// 3: DEPRECATED - not powerful enough.
// It's also no good if there is a point of embedding dimension 5.
// This is not a precise function - it may allow bad candidates
// through to be rejected later.
function point_of_embdim5(IY)
        R := Generic(IY);       // the underlying polynomial ring
        J := JacobianMatrix(Basis(IY));
        for i in [1..Rank(R)] do
                phi := set_to_zero_except_i_to_one(R,i);
                Mi := Matrix(NumberOfRows(J), NumberOfColumns(J),
                        [ phi(f) : f in Eltseq(J) ]);
                if Rank(Mi) le 1 then
                     return true;
                end if;
        end for;
        return false;
end function;
*/

// 3: Check whether a coordinate point on D cannot be a terminal singularity
// (either because it's not hyperquotient, or bcs it is but not on Mori's list.
// IY is a seq of the eqns of Y; Dindices records the indices of D coord pts.
// e.g. in P(1,1,2,2,3,3,4), if D = P(1,1,3) then Dindices = [1,2,5].
function linear_terms(f)	// f a poly; return the linear terms
	return &+[Parent(f)| t : t in Terms(f) | TotalDegree(t) eq 1 ];
end function;

// which generator is y?
whichgen := func< y | Index([R.i:i in [1..Rank(R)]],y) where R is Parent(y) >;

function non_terminal_coord_point(IY,Dindices)
	R := Universe(IY);
	gens := [ R.i : i in [1..Rank(R)] ];
	W := Grading(R);
	for i in Dindices do	// work at a coord point P_i
		// no condition if P_i not in Y
		P_i := [ 0 : j in [1..#W] ]; P_i[i] := 1;
		if &or [Evaluate(f,P_i) ne 0 : f in IY] then continue i; end if;
		if W[i] eq 1 then continue; end if;	// ignore Gorenstein pts
		phieq := gens;
		phieq[i] := 1;
		phi := hom< R -> R | phieq >;	// set i-th coord = 1
		linear := [ f1 : f in phi(IY) |
			f1 ne 0 where f1 := linear_terms(f) ];
		// use trivial gb to do gauss elimination on the coeffs
		better := GroebnerBasis(Ideal(linear));
		if #better le 1 then
			return true;	// embedding dim >= 5 so not terminal
		end if;
		// determine which weights are eliminated near P_i
		elim_inds := [ whichgen(Monomials(f)[1]) :f in better ] cat [i];
		r := W[i];
		local_wts := [ W[i] mod r : i in [1..#W] | not i in elim_inds ];
		// local_wts must include a zero (mod r) except when r is 4.
		if #local_wts eq 3 then continue i; end if;
		if not 0 in local_wts then
			if r ne 4 or Sort(local_wts) ne [1,1,2,3] then
				return true;
			else
				// don't want to, but could check here
				// whether the point is isolated.
			end if;
		else
			// don't want to, but could check here
			// whether the point is isolated also.
		end if;
	end for;
	return false;
end function;

// 4:
// Check that the singular locus lies on D, and check that it is
// reduced, and compute its length.

// The ideal of the coordinate point P_v corr to variable v in a poly ring.
function coordinate_point(v)
	RR := Generic(Parent(v));
	gens_not_v := [ RR.i : i in [1..Rank(RR)] | RR.i ne v ];
	return Ideal(gens_not_v);
end function;

// The ideal of the coordinate line L_uv corr to variables u,v in a poly ring.
function coordinate_line(L)
	u,v := Explode(L);
	RR := Generic(Parent(v));
	gens_not_uv := [ RR.i : i in [1..Rank(RR)] | not RR.i in {u,v} ];
	return Ideal(gens_not_uv);
end function;

function singular_points( M, GensID : namebase := "y", all_tests := true )
    // Build Y = Pf(M) and compute its singular (= non-quasismooth) locus.
    // We don't check it, but M has been built so that Y contains D = V(GensID).
	// tt := Cputime();
    IY := Ideal(Pfaffians(M,4));
    ID := Ideal(GensID);
    RR := Generic(IY);
    AssignNamesBase(~RR, namebase);
    othervars := [ RR.i : i in [1..Rank(RR)] | not RR.i in GensID ];
    Y := Scheme(Proj(RR),IY);
    JY := JacobianMatrix(Y);
    // Need to saturate the ideal to compare it to its radical.
    ISingY := Saturation(Ideal(Pfaffians(M,4) cat Minors(JY,3)));
	// print Cputime(tt), "Computed saturation:";
    ISingYred := Radical(ISingY);
	// print Cputime(tt), "Computed radical:";

    // Check that all singular points of Y lie on D.
    are_sings_on_D := ID subset ISingYred;
	// print Cputime(tt), "Sings lie on D: ",are_sings_on_D;
    if not all_tests and not are_sings_on_D then
        return are_sings_on_D,_,_,_;
    end if;

    // Check that the singular locus is reduced.
    sings_are_red := ISingY eq ISingYred;
	// print Cputime(tt), "Sings are reduced: ",sings_are_red;
    if not all_tests and not sings_are_red then
        return are_sings_on_D,sings_are_red,_,_;
    end if;

    // Check the 0-strata of D
    sings_on_pts := [ ISingYred subset coordinate_point(v) : v in othervars ];
	// print Cputime(tt), "Sings at 0-strata:",sings_on_pts;

    // Check the 1-strata of D
    u,v,w := Explode(othervars);
    sings_on_lines := [ ISingYred ne ColonIdeal(ISingYred, coordinate_line(L))
			 : L in [ [v,w], [u,w], [u,v] ] ];
	// print Cputime(tt), "Sings on 1-strata:",sings_on_lines;

    // Pick an affine patch of D that contains all singularities (if possible)
    smooth_line := Index(sings_on_lines, false);
    if smooth_line ne 0 then
        x := othervars[smooth_line];
    else
        error "Singularities do not all lie in a standard affine patch of D";
    end if;

    ISingD := Ideal(Pfaffians(M,4) cat Minors(JY,3) cat GensID cat [x-1]);
    RSingD := quo<RR|ISingD>;
    if HasFiniteDimension(RSingD) then
        num_sings := Dimension(RSingD); // No. of reduced pts on D with x <> 0.
    else
        num_sings := -1;
    end if;

    return are_sings_on_D,sings_are_red,sings_on_pts,sings_on_lines,num_sings;
end function;

