//////////////////////////////////////////////////////////////////////
// Records for Tom and Jerry unprojection data
// Gavin Brown, Michael Kerber, Miles Reid, 2005-10
//
//  Record formats for Fano candidates and for projections.
//
//  Printing of these records.
//	TomAndJerryPrint( A ) -> prints the Fano or unprojection record A
//
//		Notation
//		--------
//  We work with projections of the form
//	P in X in P(W) - -> E in Y in P(W omit r)
//  where
//	X	codim 4 Fano 3-fold
//	P in X	Type I centre on X of type 1/r(1,a,r-a)
//		(Type I means that r, 1, a, r-a appear separately in W)
//	Y	codim 3 - syzygy matrix determined by integer u and 5-tuple w
//	E in Y	the exceptional locus of the projection.
//
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
//	A record format for holding Tom and Jerry data
//////////////////////////////////////////////////////////////////////

// data about a family of Fanos
FanoRecord := recformat<
		// generic data about Gorenstein polarised varieties
	ID : RngIntElt,			// arbitrary integer as identifier
	weights : SeqEnum,		// ambient weights
	codim : RngIntElt,		// codimension
	basket : SeqEnum,		// e.g. [[5,2],[2,1],[2,1]]
	basket1 : SeqEnum,		// Type 1 centres <[5,2], 41=target ID>
	k : RngIntElt,			// K_X = O_X(k) if X is quasismooth
		// data about equations for codim 3 Fanos
	u : RngIntElt,
	w : SeqEnum,			// u,w determine the 5x5 weight matrix
	WM : Mtrx,			// matrix of weights of M
	M : Mtrx,			// matrix of polys defining Y
		// data of TJ projections for codim 4 Fanos
	tjequiv : SeqEnum,		// TJ equivalences e.g. T1=T2
					// written as ["Tom(1)","Tom(2)"]
	tjdata : SeqEnum,		// each proj as <[5,2],"Tom(3)",Y-id,N>
	tjall : SeqEnum >;		// all projs, including failures


// data about the projection
TomJerryRecord := recformat<
    name : MonStgElt,			// e.g. Tom(3), Jerry(4,5)
    domainID : RngIntElt,		// ID of X
    codomainID : RngIntElt,		// ID of Y
    centre : SeqEnum,			// P = [r,a]
    r : RngIntElt,			// index of P
    A : SeqEnum,			// polarisation of P, e.g. [1,a,r-a]
    Dindices : SeqEnum,			// indices of vars of ideal of D
    N : RngIntElt,			// number of nodes on D
    success : BoolElt,			// true iff unprojection can work
	// reasons why the projection might fail
    are_sings_on_D : BoolElt,		// true iff ...
    sings_on_D_reduced : BoolElt,	// true iff ...
    variable_appears : SeqEnum,		// [true iff x_i in eqns]
    pure_power_appears : SeqEnum,	// [true iff (x_i)^? in mx]
    exists_embdim5_pt : BoolElt,	// not used
    non_mori_pt_on_D : BoolElt,		// true iff a coord pt on D not terminal
    non_mori_pt_off_D : BoolElt,	// true iff a coord pt off D not term'l
    sing_is_red : BoolElt,		// true iff sing locus on D reduced
    are_0_strata_singular : SeqEnum,	// [ false iff [1,0,0] sm in D, etc.]
    are_1_strata_singular : SeqEnum,	// [ false iff [0,*,*] sm in D, etc.]
    fishy_zeroes : SeqEnum,		// seq of [i,j] where M[i,j] = 0
    collinear_zeroes : BoolElt,		// true iff collinear 0 in matrix M
    bad_zero : BoolElt,			// true iff non-isol sing from 0 in M
	// data defining the image of projection, Y
    WY : SeqEnum,			// ambient weights of Y
    u : RngIntElt,
    w : SeqEnum,			// u,w determine the 5x5 weight matrix
    WM : Mtrx,				// matrix of weights of M
    M : Mtrx,				// matrix of polys defining Y
    tjequiv : SeqEnum >;		// TJ equivalences e.g. T1=T2
					// written as ["Tom(1)","Tom(2)"]


//////////////////////////////////////////////////////////////////////
//			Printing
//////////////////////////////////////////////////////////////////////

intrinsic TomAndJerryPrint(A)
{Print the TJ object A}
	if Type(A) eq Rec and "ID" in Names(A) then // A is a candidate variety
		cod := A`codim;
		W := A`weights;
		printf "Codimension %o candidate variety Y in P%o\n",cod,W;
		if assigned A`k then
			printf "with K_Y = O(%o) and ",A`k;
		end if;
		if cod eq 3 then
			printf "with syzygy weights\n";
			print A`WM;
		end if;
	elif Type(A) eq Rec and "success" in Names(A) then // A is a projection
		works := A`success select "works" else "fails";
		printf "%o unprojection of D = P%o %o:\n",A`name,A`A,works;
		if A`success then
			pts := A`are_0_strata_singular;
			lines := A`are_1_strata_singular;
			if &or pts or &or lines then
				printf "\t%o (reduced) singularities on D\n",A`N;
				printf "\t... at 0-strata %o\n", pts;
				printf "\t... on 1-strata %o\n", lines;
			else
				printf "\t%o (reduced) singularities on big stratum of D\n",A`N;
			end if;
			if assigned A`fishy_zeroes and #A`fishy_zeroes ne 0 then
				printf "\tMatrix zeroes at ";
				for Z in A`fishy_zeroes do
					printf "[%o,%o] ",Z[1],Z[2];
				end for;
				printf "\n";
			end if;
		else
			if assigned A`collinear_zeroes and A`collinear_zeroes then
				printf "\tcollinear zeroes ";
				for Z in A`fishy_zeroes do
					printf "[%o,%o] ",Z[1],Z[2];
				end for;
				printf "in the matrix so not irreducible\n";
			end if;
			if assigned A`variable_appears then
				vars := A`variable_appears;
				if not &and vars then
					printf "\tmissing variables: %o\n",
						[i:i in [1..#vars]|not vars[i]];
				end if;
			end if;
			if assigned A`pure_power_appears then
				vars := A`pure_power_appears;
				if not &and vars then
				    printf "\tcoord points of emb dim 6: %o\n",
						[i:i in [1..#vars]|not vars[i]];
				end if;
			end if;
			if assigned A`non_mori_pt_on_D and A`non_mori_pt_on_D then
				printf "\tthere is a nonterminal coordinate point on D\n";
			end if;
			if assigned A`non_mori_pt_off_D and A`non_mori_pt_off_D then
				printf "\tthere is a nonterminal coordinate point off D\n";
			end if;
			if assigned A`sing_is_red and not A`sing_is_red then
				printf "\tsingular locus is not reduced\n";
			end if;
			if assigned A`N and A`N eq -1 then
				printf "\tsingular locus is not isolated\n";
			end if;
		end if;
	elif Type(A) eq SeqEnum then	// assume it's a seq of unprojections
		for p in A do
			if p`success then
				TomAndJerryPrint(p);
			end if;
		end for;
	else	// give up - if you add new record types, include printing here
		A;
	end if;
end intrinsic;

